local Dongle = DongleStub("Dongle-1.0")

WheresMyCow = Dongle:New("WheresMyCow")

Cow = WheresMyCow

Cow.Addons = {}

Cow.Friesian = CreateFrame("Button", "CowShift",UIParent)
Cow.Jersey   = CreateFrame("Button", "CowEnter", UIParent)

-- Pool of tables to reuse.
Cow.tablePool = setmetatable({}, 
				( {__index = function(self, n)
		        				if n==0 then
	    							return table.remove(self) or {}
								end
                            end})
						)

local function getTable(...)
	local tab = Cow.tablePool[0]
	if select("#", ...) > 0 then
		for i=1,select("#", ...) do
			table.insert(tab, (select(i, ...)))
		end
	end
	return tab
end

local function returnTable(tab)
	for i in pairs(tab) do
		tab[i] = nil
	end
	table.insert(Cow.tablePool, tab)
end

function Cow:TableStats()
	self:Print("Pool Size:"..#Cow.tablePool)
end

local attributes = { "state", "showstates", "hidestates", "newstate" }

function Cow:Initialize()
	self.cmds = self:InitializeSlashCommand("Where's My Cow? - v1.0", "wmc", "mooo", "wmc", "cow")
	self.cmds:RegisterSlashHandler("Activate Where's My Cow Mode.", "^(%w+)$", "ToggleScanning")
    self.uncmds = self:InitializeSlashCommand("Where's My Cow? - v1.0", "unwmc", "uncow")
    self.uncmds:RegisterSlashHandler("Disable Where's My Cow Mode.", "", function() self:ToggleScanning(self.Scanning) end)
	self.Friesian:SetScript("OnUpdate",function() WheresMyCow:Update() end)
    self.Friesian:SetScript("OnClick", function() WheresMyCow:TabCycle() end)
    self.Jersey:SetScript("OnClick", function() WheresMyCow:ChooseFrame() end)
    self.Tab = 0
    self.TableAdded = 0
end

-- Pass in the name of the Addon and the Addon table
function Cow:RegisterAddon(AddonName) 
	if not self.Addons[AddonName] then 
		self.Addons[AddonName] = {Frames = getTable(), RegisterFrame = function(self, Frame) WheresMyCow:RegisterFrame(self, Frame) end}
	end
	return self.Addons[AddonName]
end

-- Run through the entire addon, storing table references
function Cow:RegisterFrame(Addon, Frame, Info)
	if Frame.GetObjectType and Frame:IsObjectType("frame") then
		Addon.Frames[Frame] = {Info = Info, Name= Frame:GetName(), State = getTable()}
	end
end

function Cow:PrintRegisteredTables(AddonName)
	if not self.Addons[AddonName] then return end
	for i,v in pairs(self.Addons[AddonName].Frames) do
		self:PrintF(v.Name)
	end
end

function Cow:ToggleScanning(Addon)
    if not Addon then return end
	if not self.Addons[Addon] then
		self:PrintF("That goes 'Hrurrruugh!'. It is a %s. That's not my Cow.", Addon)
		return
	end
	if self.Scanning then
		self.Scanning = nil
        SetBinding("SHIFT-TAB", self.STabBinding)
        SetBinding("SHIFT-ENTER", self.EnterBinding)
		self.Friesian:Hide()
		self:PrintF("That's Not My %s!",Addon)
	else
		self:PrintF("Where's My %s? Is That My %s?",Addon, Addon)
		self.Scanning = Addon
        self.STabBinding = GetBindingAction("SHIFT-TAB")
        self.EnterBinding = GetBindingAction("SHIFT-ENTER")
        SetBindingClick("SHIFT-TAB", "CowShift")
        SetBindingClick("SHIFT-ENTER", "CowEnter")
		self.Friesian:Show()
	end
	return
end

function Cow:Update()
	if self.Scanning and IsModifierKeyDown() then
		local x,y = GetCursorPosition()
        local info =  getTable()
		for f,v in pairs(self.Addons[self.Scanning].Frames) do
			local es = f:GetEffectiveScale()
			if  (x >= (f:GetLeft() or -1)*es)	and (x <= (f:GetRight() or -2)*es) and
				(y >= (f:GetBottom() or -1)*es) and (y <= (f:GetTop() or -2)*es) then
                table.insert(info, getTable(f,v))
			end
		end
        if #info > 0 then
            if self.SelectedFrame then
                local frame = getTable(info[self.SelectedFrame + 1])
                self:GenerateTooltip(frame)
                returnTable(info)
                returnTable(frame)
                return
            end
            self:GenerateTooltip(info)
        end
        returnTable(info)
    elseif self.Scanning and GameTooltip:IsShown() then 
        GameTooltip:Hide()
	end
end

function Cow:TabCycle()
    Cow.Tab = 1 + Cow.Tab
end

function Cow:ChooseFrame()
    if self.SelectedFrame then
        self.SelectedFrame = nil
    else
        self.SelectedFrame = Cow.Tab
    end
end

function Cow:GenerateTooltip(frames)
    local info = getTable()
    local tt = GameTooltip
    tt:SetOwner(Cow.Friesian, "ANCHOR_CURSOR")
    tt:AddLine("Where's My Cow?")
    Cow:AddToTooltip(tt, "Key", "Value")
    
    if not self.SelectedFrame then 
        Cow.Tab = math.fmod(Cow.Tab, #frames)
    end

    for i,d in pairs(frames) do
    	if tt:NumLines() > 28 then 
            tt:AddLine("Truncated")
    		break;
    	end
        local v = d[2]
        if not self.SelectedFrame and Cow.Tab == i - 1 then
            Cow:AddToTooltip(tt, "Name", ">"..v.Name.."<")
        else
            Cow:AddToTooltip(tt, "Name", v.Name)
        end

        if self.SelectedFrame or #frames == 1 then
            self:ShowDetails(tt,d)
        end

    	if v.Info and #v.Info > 0 and #v.Info + tt:NumLines() < 30 then
    		for i,v in pairs(v.Info) do
    			Cow:AddToTooltip(tt, v[1], v[2])
    		end
    	elseif v.Info and #v.Info > 0 then
            tt:AddLine("Truncated...")
    	end
    end
    GameTooltip:Show() 
end

function Cow:ShowDetails(tt, frame)
    -- Show any securestate information
    if frame[1]:IsProtected() then
        for i,v in ipairs(attributes) do
            if frame[1]:GetAttribute(v) then
                Cow:AddToTooltip(tt, v, frame[1]:GetAttribute(v))
            end
        end
    end

    -- Show the children
    if frame[1].GetChildren and frame[1]:GetChildren() then
        local kids = ""
        for i=1, select("#", frame[1]:GetChildren()) do
            kids = kids..select(i, frame[1]:GetChildren()):GetName().. ", "
        end
        if kids ~= "" then
            Cow:AddToTooltip(tt, "Children:", kids)
        end
    end
    -- Show the parent
    if frame[1]:GetParent() then
        Cow:AddToTooltip(tt, "Parent:", frame[1]:GetParent():GetName())
    end
end

function Cow:AddToTooltip(tt, left, right)
    tt:AddDoubleLine(left, right, 0.5, 0.5, 1, 1, 1 ,1)
end
